'''
generate.py

The core generation code and main function of the Infinite Mario ORE generator.

Copyright 2010 Peter Mawhorter
This software is distributed under the terms of the GNU General Public License.
'''

import random, sys

import curate

import ore.chunk
import ore.library
import ore.state
import ore.position

lib = ore.library.Library(set(curate.load("res/chunks.lib")))

width, height = 130, 12

min_height = 3
reserved_space = 13

base = curate.unpack(
  '''\
Tags:
start
Map:
      
      
      
      
      
      
      
      
    ` 
++++++
++++++
++++++
Enemies:
      
      
      
      
      
      
      
      
      
      
      
      
End\
'''
)
finish = curate.unpack(
  '''\
Tags:
finish
Map:
              
              
              
              
              
              
              
              
 ` X          
++++++++++++++
++++++++++++++
++++++++++++++
Enemies:
              
              
              
              
              
              
              
              
   X          
++++++++++++++
++++++++++++++
++++++++++++++
End\
'''
)

base.add_chunk(
  finish,
  take_relative_to = ore.position.GridPosition(2, 8),
  put_relative_to = ore.position.GridPosition(width - reserved_space, 8),
)

def anchor_metric(anchor, state):
  if anchor in state.exhausted:
    return 0
  if not (0 <= anchor.x < width) or not (min_height <= anchor.y < height):
    return 0
  return 0.01 ** state.memories["boredom"][anchor]

def chunk_metric(chunk, anchored_at, state):
  for pos in chunk.coverage:
    dst = pos - anchored_at + state.current_anchor
    if dst.x < 0\
    or dst.x > width - reserved_space\
    or dst.y < min_height\
    or dst.y > height:
      return 0
  result = 1;
  result *= 0.7 ** state.memories["boredom"][chunk]

  return result

generator = ore.state.State(
  lib,
  base,
  anchor_metric=anchor_metric,
  chunk_metric=chunk_metric,
  memory_initializer=ore.state.memories_of_boredom,
  memory_creator=ore.state.remember_boredom,
)

def completed(state):
  cv = { (pos.x, pos.y) for pos in state.result.coverage }
  gw = 0
  hit = False
  for x in range(width):
    hit = False
    for y in range(height, -1, -1):
      if (x, y) in cv:
        hit = True
        break
    if hit:
      gw = 0
    else:
      gw += 1
    if gw > 5:
      return False
  return True

try:
  generator.converge(completed, check_every=1)
except KeyboardInterrupt as e:
  sys.excepthook(*sys.exc_info())

curate.save([generator.result], 'result.lvl')
