'''
hierarchy.py

Implements a hierarchy of ComponentTypes for the InfiniteMario domain.

Copyright 2010 Peter Mawhorter
This software is distributed under the terms of the GNU General Public License.
'''

from ore.component import ComponentType as CT

platform = CT("platform", {"map", "solid"})
enemy = CT("enemy", {"enemy", "solid"})
winged_enemy = CT("winged_enemy", {"enemy", "solid"}, {enemy})
special_enemy = CT("special_enemy", {"enemy", "solid"})
decoration = CT("decoration", {"map"})
edge = CT("edge", None)
exit = CT("exit", {"map"})
coin = CT("coin", {"map"})

ledge = CT("ledge", {"map", "solid"}, {platform})
ledge_fill = CT("ledge_fill", {"map"}, {platform})

ground = CT("ground", {"map", "solid"}, {platform})
block = CT("block", {"map", "solid"}, {platform})

normal_ground = CT("normal_ground", {"map", "solid"}, {ground})
pipe = CT("pipe", {"map", "solid"}, {ground})
skinny_pipe = CT("skinny_pipe", {"map", "solid"}, {ground})

stacked_logs = CT("stacked_logs", {"map", "solid"}, {ground})

bullet_canon = CT("bullet_canon", {"map", "solid"}, {platform, special_enemy})
bullet_tower = CT("bullet_tower", {"map", "solid"}, {platform})

question_block = CT("question_block", {"map", "solid"}, {block})
brick_block = CT("brick_block", {"map", "solid"}, {block})
flip_block = CT("flip_block", {"map", "solid"}, {block})
rock_block = CT("rock_block", {"map", "solid"}, {block})
wood_block = CT("wood_block", {"map", "solid"}, {block})
blue_block = CT("blue_block", {"map", "solid"}, {block})

empty_question = CT("empty_question", {"map", "solid"}, {question_block})
coin_question = CT("coin_question", {"map", "solid"}, {question_block})
powerup_question = CT("powerup_question", {"map", "solid"}, {question_block})

empty_brick = CT("empty_brick", {"map", "solid"}, {brick_block})
coin_brick = CT("coin_brick", {"map", "solid"}, {brick_block})
powerup_brick = CT("powerup_brick", {"map", "solid"}, {brick_block})

goomba = CT("goomba", {"enemy", "solid"}, {enemy})
green_koopa = CT("green_koopa", {"enemy", "solid"}, {enemy})
red_koopa = CT("red_koopa", {"enemy", "solid"}, {enemy})
spike_shell = CT("spike_shell", {"enemy", "solid"}, {enemy})

winged_goomba = CT("winged_goomba", {"enemy", "solid"}, {winged_enemy})
winged_green_koopa = CT("winged_green_koopa", {"enemy", "solid"},{winged_enemy})
winged_red_koopa = CT("winged_red_koopa", {"enemy", "solid"}, {winged_enemy})
winged_spike_shell = CT("winged_spike_shell", {"enemy", "solid"},{winged_enemy})

piranha_flower = CT("piranha_flower", {"enemy"}, {special_enemy})

arrow_sign = CT("arrow_sign", {"map"}, {decoration})
grass = CT("grass", {"map"}, {decoration})
bush = CT("bush", {"map"}, {decoration})


type_to_char = {
  coin: '$',
  exit: 'X',
  ledge: '-',
  ledge_fill: ')',
  normal_ground: '+',
  pipe: '[',
  skinny_pipe: '|',
  stacked_logs: 'o',
  bullet_canon: '%',
  bullet_tower: ']',
  flip_block: '/',
  rock_block: '.',
  wood_block: ';',
  blue_block: ':',
  empty_question: '&',
  coin_question: '?',
  powerup_question: '!',
  empty_brick: '#',
  coin_brick: '@',
  powerup_brick: '*',
  goomba: 'g',
  green_koopa: 'k',
  red_koopa: 'r',
  spike_shell: 's',
  winged_goomba: 'G',
  winged_green_koopa: 'K',
  winged_red_koopa: 'R',
  winged_spike_shell: 'S',
  piranha_flower: 'f',
  arrow_sign: '>',
  grass: ',',
  bush: '~',
}

char_to_index = {
  '-': 4 + 1 + 8*16,
  ')': 4 + 1 + 9*16,
  '+': 0 + 1 + 9*16,
  '[': 8 + 2 + 0*16,
  '|': 8 + 0 + 2*16,
  'o': 8 + 1 + 2*16,
  '%': 12 + 2 + 0*16,
  ']': 12 + 2 + 2*16,
  '/': 4 + 0 + 2*16,
  '.': 8 + 1 + 0*16,
  ';': 12 + 0 + 0*16,
  ':': 12 + 0 + 1*16,
  '&': 4 + 0 + 0*16,
  '?': 4 + 1 + 1*16,
  '!': 4 + 2 + 1*16,
  '#': 0 + 0 + 1*16,
  '@': 0 + 1 + 1*16,
  '*': 0 + 2 + 1*16,
  'g': (2, False),
  'k': (1, False),
  'r': (0, False),
  's': (3, False),
  'G': (2, True),
  'K': (1, True),
  'R': (0, True),
  'S': (3, True),
  'f': (4, False),
  '>': 0 + 3 + 4*16,
  ',': 0 + 1 + 5*16,
  '~': 0 + 1 + 6*16,
}

char_to_type = {char: type for (type, char) in type_to_char.items()}
index_to_char = {index: char for (char, index) in char_to_index.items()}

# Some extra mappings (see mapsheet.png from the res/ folder):
for x in range(4):
  index_to_char[x + 2*16] = '$'

for x in range(4, 7):
  index_to_char[x + 8*16] = '-'
  index_to_char[x + 11*16] = '-'

for x in range(4, 7):
  for y in range(9, 11):
    index_to_char[x + y*16] = ')'

for y in range(8, 12):
  index_to_char[4 + 3 + y*16] = ')'

for x in range(4):
  for y in range(8, 12):
    index_to_char[x + y*16] = '+'

for x in range(8, 16):
  for y in range(8, 12):
    if y != 11 or x in (11, 15):
      index_to_char[x + y*16] = '+'

for x in range(10, 12):
  for y in range(2):
    index_to_char[x + y*16] = '['

for y in range(1, 4):
  index_to_char[8 + y*16] = '|'

for y in range(1, 4):
  index_to_char[9 + y*16] = 'o'

for y in range(1, 3):
  index_to_char[14 + y*16] = ']'

for x in range(4, 8):
  index_to_char[x + 2*16] = '/'

index_to_char[4 + 0 + 1*16] = '?'
index_to_char[4 + 3 + 1*16] = '?'
index_to_char[0 + 3 + 1*16] = '#'

for x in range(3):
  index_to_char[x + 5*16] = ','

for x in range(3):
  index_to_char[x + 6*16] = '~'
