'''
position.py

Code that defines abstract positions. A concrete extension of this should be
used in systems that use ORE. A grid position class is supplied here as both
a reference and a useful tool.

Copyright 2010 Peter Mawhorter
This software is distributed under the terms of the GNU General Public License.
'''

class Position:

  # The abstract origin (filled in later):
  origin = None

  def __init__(self):
    pass

  def __str__(self):
    return "the origin"

  def __repr__(self):
    return "Position()"

  def __add__(self, other):
    if isinstance(other, Position):
      return other
    else:
      return NotImplemented

  def __sub__(self, other):
    if isinstance(other, Position):
      return other
    else:
      return NotImplemented

  def __radd__(self, other):
    if isinstance(other, Position):
      return other
    else:
      return NotImplemented

  def __rsub__(self, other):
    if isinstance(other, Position):
      return other
    else:
      return NotImplemented

  def __eq__(self, other):
    return type(other) == Position

  def __ne__(self, other):
    return type(other) != Position

  def __hash__(self):
    return 178179180

# Fill in the origin (the class had to be defined first):
Position.origin = Position()

origin = Position.origin

class GridPosition(Position):
  '''
  A grid position is an integer point on a 2-dimensional grid.
  '''

  # We'll fill this in below. Note that in many cases, Position.origin can be
  # used just as well.
  origin = None

  @property
  def x(self):
    return self._x

  @property
  def y(self):
    return self._y

  def __init__(self, x=0, y=0):
    self._x = x
    self._y = y

  def __str__(self):
    return "({}, {})".format(self._x, self._y)

  def __repr__(self):
    return "GridPosition({}, {})".format(self._x, self._y)

  def __add__(self, other):
    if isinstance(other, GridPosition):
      return GridPosition(self._x + other._x, self._y + other._y)
    else:
      return NotImplemented

  def __sub__(self, other):
    if isinstance(other, GridPosition):
      return GridPosition(self._x - other._x, self._y - other._y)
    else:
      return NotImplemented

  def __eq__(self, other):
    return isinstance(other, GridPosition)\
      and self._x == other._x\
      and self._y == other._y

  def __ne__(self, other):
    return not self.__eq__(other)

  def __hash__(self):
    return ((self._x + self._y) * self._y) | self._x

# Fill in the origin (the class had to be defined before we could do this):
GridPosition.origin = GridPosition(0, 0)
