'''
generate.py

The core generation code and main function of the Wesnoth ORE generator.

Copyright 2010 Peter Mawhorter
This software is distributed under the terms of the GNU General Public License.
'''

import random

import curate
import hierarchy
import metrics

import ore.library
import ore.state

lib = ore.library.Library(set(curate.load("res/chunks.lib")))

height = 15
width = 30

startx = random.choice(range(5, width//2 - 6, 2))
starty = random.randint(3, 8)
if random.randint(0, 1):
  startx = width - startx

base = curate.unpack(
  '''\
Chunk:
Tags:
unique
odd_column
Offset:
{x}, {y}
Map:
   , Ww, *Ss, Ss,   
*Rb, Ce,  Ce, Ce,*Gg
 Gg, Ce,1 Ke, Ce, Gg
   , Gg,  Ce, Gg,   
   ,*Rr,  Rr,   ,   
End\
'''.format(x = startx, y = starty)
)

generator = ore.state.State(
  lib,
  base,
  anchor_metric=metrics.build_anchor_metric(width, height),
  chunk_metric=metrics.build_chunk_metric(width, height),
  memory_initializer=metrics.init_memory,
  memory_creator=metrics.record_memory,
)

try:
  generator.converge(
    metrics.build_completion_metric(width, height),
    check_every=5
  )
except KeyboardInterrupt:
  pass

curate.save([generator.result], 'result.lib')

reflected = ore.chunk.Chunk()
for pos in generator.result.coverage:
  rpos = ore.position.GridPosition(width - pos.x, 2*height - pos.y)
  # Even/odd shift to preserve reflected form on the hex grid:
  if rpos.x % 2:
    rpos = ore.position.GridPosition(rpos.x, rpos.y + 1)
  for component in generator.result.get(pos):
    refl = component.at(rpos)
    reflected.add_component(component)
    reflected.add_component(refl)

# Fix starting positions:
maxsp = 0
for c in sorted(
  [
    _ for _ in reflected.components
    if _.is_instance_of(hierarchy.starting_position)
  ],
  key=lambda c: hierarchy.sp_to_nr[c.real_type]
):
  n = hierarchy.sp_to_nr[c.real_type]
  if n > maxsp:
    maxsp = n
  else:
    maxsp += 1
    reflected.remove_component(c)
    reflected.add_component(
      ore.component.Component(
        hierarchy.nr_to_sp[maxsp],
        c.position
      )
    )

curate.build_map(reflected, 'result.map', width=width, height=height*2)
