'''
hierarchy.py

Implements a hierarchy of ComponentTypes for the Wesnoth domain.

Copyright 2010 Peter Mawhorter
This software is distributed under the terms of the GNU General Public License.
'''

from ore.component import ComponentType as CT
from ore.component import Component as C
from ore.position import origin

starting_position = CT("starting_position", {"start_position"})

nr_to_sp = {}
for i in range(1, 10):
  exec(
"""\
sp{n} = CT("sp{n}", {{"start_position"}}, {{starting_position}})
nr_to_sp[{n}] = sp{n}
""".format(n=i)
)
sp_to_nr = {sp: nr for (nr, sp) in nr_to_sp.items()}

##########################
# The root for terrains: #
##########################

any_type = CT("any_type", set())

#########################
# Movement type groups: #
#########################

movement_category = CT("movement_category", {any_type})

move_deep_water = CT("move_deep_water", set(), {movement_category})
move_shallow_water = CT("move_shallow_water", set(), {movement_category})
move_reef = CT("move_reef", set(), {movement_category})
move_swamp_water = CT("move_swamp_water", set(), {movement_category})
move_flat = CT("move_flat", set(), {movement_category})
move_sand = CT("move_sand", set(), {movement_category})
move_forest = CT("move_forest", set(), {movement_category})
move_hills = CT("move_hills", set(), {movement_category})
move_mountains = CT("move_mountains", set(), {movement_category})
move_village = CT("move_village", set(), {movement_category})
move_castle = CT("move_castle", set(), {movement_category})
move_cave = CT("move_cave", set(), {movement_category})
move_frozen = CT("move_frozen", set(), {movement_category})
move_fungus = CT("move_fungus", set(), {movement_category})
move_none = CT("move_none", set(), {movement_category})
move_unwalkable = CT("move_unwalkable", set(), {movement_category})

###########################
# Conceptual type groups: #
###########################

conceptual_category = CT("conceptual_category", set(), {any_type})

arid = CT("arid", {"base"}, {conceptual_category})
winter = CT("winter", {"base"}, {conceptual_category})
wet = CT("wet", {"base"}, {conceptual_category})
underground = CT("underground", {"base"}, {conceptual_category})
interior = CT("interior", {"base"}, {conceptual_category})
verdant = CT("verdant", {"overlay"}, {conceptual_category})
turning = CT("turning", {"overlay"}, {conceptual_category})
dormant = CT("dormant", {"overlay"}, {conceptual_category})
tropical = CT("tropical", {"overlay"}, {conceptual_category})
earthen = CT("earthen", {"base"}, {conceptual_category})
hewn = CT("hewn", {"base"}, {conceptual_category})
dry = CT("dry", {"base"}, {conceptual_category})
improvement = CT("improvement", {"overlay"}, {move_flat, conceptual_category})
primitive = CT("primitive", {"base", "overlay"}, {conceptual_category})
ruin = CT("ruin", {"base", "overlay"}, {conceptual_category})
human = CT("human", {"base", "overlay"}, {conceptual_category})
orcish = CT("orcish", {"base", "overlay"}, {conceptual_category})
elven = CT("elven", {"base", "overlay"}, {conceptual_category})
dwarven = CT("dwarven", {"base", "overlay"}, {conceptual_category})
drake = CT("drake", {"base", "overlay"}, {conceptual_category})
merfolk = CT("merfolk", {"base", "overlay"}, {conceptual_category})
lr = CT("lr", {"overlay"}, {conceptual_category})
rl = CT("rl", {"overlay"}, {conceptual_category})
c = CT("c", {"overlay"}, {conceptual_category})

###########################
# Mechanical type groups: #
###########################

mechanics_category = CT("mechanics_category", set(), {any_type})

water = CT("water", {"base"}, {wet, mechanics_category})
frozen = CT("frozen", {"base"}, {winter, move_frozen, mechanics_category})
heals = CT("heals", {"overlay"}, {mechanics_category})
income = CT("income", {"overlay"}, {mechanics_category})
forested = CT("forested", {"overlay"}, {mechanics_category})
flat = CT("flat", {"base"}, {move_flat, mechanics_category})
sandy = CT("sandy", {"base"}, {move_sand, mechanics_category})
cave = CT("cave", {"base"}, {underground, move_cave, mechanics_category})
rough = CT("rough", {"base"}, {move_hills, mechanics_category})
mountainous = CT("mountainous", {"base"}, {move_mountains, mechanics_category})
unwalkable = CT(
  "unwalkable",
  {"base", "overlay"},
  {move_unwalkable, mechanics_category}
)
impassable = CT(
  "impassable",
  {"base", "overlay"},
  {unwalkable, move_none, mechanics_category}
)
recruit_into = CT("recruit_into", {"base"}, {move_castle, mechanics_category})
recruit_from = CT("recruit_from", {"base"}, {move_castle, mechanics_category})
decoration = CT("decoration", {"overlay"}, {mechanics_category})
illuminated = CT("illuminated", {"overlay"}, {mechanics_category})
molten = CT(
  "molten",
  {"base"},
  {illuminated, move_unwalkable, mechanics_category}
)

##################
# Terrain types: #
##################

# Water:
swamp = CT("swamp", {"base"}, {water, move_swamp_water})
shallow_water = CT("shallow_water", {"base"}, {water, move_shallow_water})
deep_water = CT("deep_water", {"base"}, {water, move_deep_water})
ford = CT("ford", {"base"}, {water, flat, move_shallow_water})
reef = CT("reef", {"base"}, {water, move_reef})

# Flat:
grass = CT("grass", {"base"}, {flat})
dirt = CT("dirt", {"base"}, {flat})
cobbles = CT("cobbles", {"base"}, {flat})

# Icy:
ice = CT("ice", {"base"}, {frozen})
snow = CT("snow", {"base"}, {frozen})

# Arid:
desert = CT("desert", {"base"}, {arid, sandy})
sand = CT("sand", {"base"}, {arid, sandy})
oasis = CT("oasis", {"base", "overlay"}, {arid, heals, sandy})
rubble = CT("rubble", {"overlay"}, {arid, rough})
crater = CT("crater", {"base", "overlay"}, {arid, flat, sandy})

# Forest (see the flavors section for more):
forest = CT("forest", {"overlay"}, {forested, move_forest})

# Conceptual forest sub-groups:
deciduous = CT("deciduous", {"overlay"}, {forest})
evergreen = CT("evergreen", {"overlay"}, {forest})

# Hills:
hills = CT("hills", {"overlay"}, {rough})

# Mountains:
mountains = CT("mountains", {"base"}, {mountainous})

# Interior:
wood_floor = CT("wood_floor", {"base"}, {flat, interior})

# Underground:
light_beam = CT("light_beam", {"overlay"}, {underground, illuminated})
cave_floor = CT("cave_floor", {"base"}, {cave})
cave_path = CT("cave_path", {"base"}, {underground, flat})
flagstones = CT("flagstones", {"base"}, {underground, flat})
mushroom_grove = CT(
  "mushroom_grove",
  {"overlay"},
  {underground, forested, move_fungus}
)
rockbound_cave = CT("rockbound_cave", {"base"}, {cave, rough})

# Impassable:
chasm = CT("chasm", {"base"}, {unwalkable})
lava = CT("lava", {"base"}, {unwalkable, molten})
volcano = CT("volcano", {"base"}, {unwalkable, mountainous, molten})
impassable_mountains = CT(
  "impassable_mountains",
  {"base", "overlay"},
  {impassable, mountainous}
)
cave_wall = CT("cave_wall", {"base", "overlay"}, {impassable, underground})
stone_wall = CT("stone_wall", {"base", "overlay"}, {impassable, interior})
impassibility = CT("impassibility", {"overlay"}, {impassable})
void = CT("void", {"base", "overlay"}, {impassable})

# Villages:
village = CT("village", {"overlay"}, {move_village, heals, income})

# Castles:
encampment = CT("encampment", {"overlay"}, {recruit_into, primitive})
castle = CT("castle", {"overlay"}, {recruit_into})

# Keeps:
bastion = CT("bastion", {"overlay"}, {recruit_from, primitive})
keep = CT("keep", {"overlay"}, {recruit_from})

# Improvements:
tracks = CT("tracks", {"overlay"}, {improvement})
wood_bridge = CT("wood_bridge", {"overlay"}, {improvement})
stone_bridge = CT("stone_bridge", {"overlay"}, {improvement})
chasm_bridge = CT("chasm_bridge", {"overlay"}, {improvement})


# Decorations:
flowers = CT("flowers", {"overlay"}, {decoration})
crops = CT("crops", {"overlay"}, {decoration})
stones = CT("stones", {"overlay"}, {decoration})
mushrooms = CT("mushrooms", {"overlay"}, {decoration})
mushroom_farm = CT("mushroom_farm", {"overlay"}, {decoration, underground})
desert_plants = CT("desert_plants", {"overlay"}, {decoration, arid})
windmill = CT("windmill", {"overlay"}, {decoration})

############
# Flavors: #
############

# Water:
grey_shallow_water = CT("grey_shallow_water", {"base"}, {shallow_water})
grey_deep_water = CT("grey_deep_water", {"base"}, {deep_water})
tropical_shallow_water = CT(
  "tropical_shallow_water",
  {"base"},
  {tropical, shallow_water}
)
tropical_deep_water = CT(
  "tropical_deep_water",
  {"base"},
  {tropical, deep_water}
)
mud = CT("mud", {"base"}, {swamp})

# Flat:
semi_dry_grass = CT("semi_dry_grass", {"base"}, {dry, grass})
dry_grass = CT("dry_grass", {"base"}, {dry, grass})
leaf_litter = CT("leaf_litter", {"base"}, {grass})
dark_dirt = CT("dark_dirt", {"base"}, {dirt})
dry_dirt = CT("dry_dirt", {"base"}, {arid, dry, dirt})
clean_cobbles = CT("clean_cobbles", {"base"}, {cobbles})
overgrown_cobbles = CT("overgrown_cobbles", {"base"}, {cobbles})
alt_desert_plants = CT("alt_desert_plants", {"overlay"}, {desert_plants})

# Forest:
great_tree = CT("great_tree", {"overlay"}, {forest})
deciduous_forest = CT("deciduous_forest", {"overlay"}, {deciduous})
evergreen_forest = CT("evergreen_forest", {"overlay"}, {evergreen})
tropical_forest = CT("tropical_forest", {"overlay"}, {tropical, forest})
mixed_forest = CT(
  "mixed_forest",
  {"overlay"},
  {deciduous_forest, evergreen_forest}
)
deciduous_forest_summer = CT(
  "deciduous_forest_summer",
  {"overlay"},
  {deciduous_forest, verdant}
)
deciduous_forest_fall = CT(
  "deciduous_forest_fall",
  {"overlay"},
  {deciduous_forest, turning}
)
deciduous_forest_winter = CT(
  "deciduous_forest_winter",
  {"overlay"},
  {deciduous_forest, dormant}
)
snow_deciduous_forest = CT(
  "snow_deciduous_forest",
  {"overlay"},
  {deciduous_forest_winter, frozen}
)
snow_evergreen_forest = CT(
  "snow_evergreen_forest",
  {"overlay"},
  {evergreen_forest, frozen}
)
mixed_forest_summer = CT(
  "mixed_forest_summer",
  {"overlay"},
  {mixed_forest, verdant}
)
mixed_forest_fall = CT(
  "mixed_forest_fall",
  {"overlay"},
  {mixed_forest, turning}
)
mixed_forest_winter = CT(
  "mixed_forest_winter",
  {"overlay"},
  {mixed_forest, dormant}
)
snow_mixed_forest = CT(
  "snow_mixed_forest",
  {"overlay"},
  {mixed_forest_winter, frozen}
)

# Rough:
dry_hills = CT("dry_hills", {"base"}, {dry, hills})
dunes = CT("dunes", {"base"}, {arid, hills})
snow_hills = CT("snow_hills", {"base"}, {frozen, hills})

# Mountains:
dry_mountains = CT("dry_mountains", {"base"}, {dry, mountains})
snow_mountains = CT("snow_mountains", {"base"}, {frozen, mountains})

# Underground:
earthen_floor = CT("earthen_floor", {"base"}, {earthen, cave_floor})
illuminated_mushroom_grove = CT(
  "illuminated_mushroom_grove",
  {"overlay"},
  {mushroom_grove, illuminated}
)

# Obstacles:
earthen_chasm = CT("earthen_chasm", {"base"}, {earthen, chasm})
ethereal_abyss = CT("ethereal_abyss", {"base"}, {chasm})
lava_chasm = CT("lava_chasm", {"base"}, {lava, chasm})
dry_impassable_mountains = CT(
  "dry_impassable_mountains",
  {"base", "overlay"},
  {dry, impassable_mountains}
)
snow_impassable_mountains = CT(
  "snow_impassable_mountains",
  {"base", "overlay"},
  {frozen, impassable_mountains}
)
hewn_cave_wall = CT("hewn_cave_wall", {"base", "overlay"}, {cave_wall, hewn})
earthen_cave_wall = CT(
  "earthen_cave_wall",
  {"base", "overlay"},
  {cave_wall, earthen}
)
hewn_earthen_cave_wall = CT(
  "hewn_earthen_cave_wall",
  {"base", "overlay"},
  {cave_wall, hewn, earthen}
)
lit_stone_wall = CT("lit_stone_wall", {"base", "overlay"}, {stone_wall})

# Villages:
adobe_village = CT("adobe_village", {"overlay"}, {village})
desert_tents = CT("desert_tents", {"overlay"}, {village})
tents = CT("tents", {"overlay"}, {village})
orcish_village = CT("orcish_village", {"overlay"}, {orcish, village})
snow_orcish_village = CT(
  "snow_orcish_village",
  {"overlay"},
  {winter, orcish_village}
)
elven_village = CT("elven_village", {"overlay"}, {elven, village})
snow_elven_village = CT(
  "snow_elven_village",
  {"overlay"},
  {winter, elven_village}
)
cottage = CT("cottage_village", {"overlay"}, {village})
snow_cottage = CT("snow_cottage_village", {"overlay"}, {winter, cottage})
city_building = CT("city_building", {"overlay"}, {village})
snow_city_building = CT(
  "snow_city_building",
  {"overlay"},
  {winter, city_building}
)
hill_house = CT("hill_house", {"overlay"}, {village})
snow_hill_house = CT("snow_hill_house", {"overlay"}, {winter, hill_house})
tropical_village = CT("tropical_village", {"overlay"}, {tropical, village})
drake_village = CT("drake_village", {"overlay"}, {drake, village})
cave_village = CT("cave_village", {"overlay"}, {underground, village})
dwarven_village = CT(
  "dwarven_village",
  {"overlay"},
  {underground, dwarven, village}
)
hut = CT("hut", {"overlay"}, {village})
snow_hut = CT("snow_hut", {"overlay"}, {winter, hut})
log_cabin = CT("log_cabin", {"overlay"}, {village})
snow_log_cabin = CT("snow_log_cabin", {"overlay"}, {winter, log_cabin})
swamp_village = CT("swamp_village", {"overlay"}, {wet, village})
merfolk_village = CT(
  "merfolk_village",
  {"base", "overlay"},
  {shallow_water, merfolk, village}
)
swamp_merfolk_village = CT(
  "merfolk_village",
  {"base", "overlay"},
  {swamp, merfolk, village}
)

# Castles:
human_fort = CT("human_fort", {"base", "overlay"}, {human, encampment})
snow_human_fort = CT(
  "snow_human_fort",
  {"base", "overlay"},
  {winter, human_fort}
)
orcish_fort = CT("orcish_fort", {"base", "overlay"}, {orcish, encampment})
snow_orcish_fort = CT(
  "snow_orcish_fort",
  {"base", "overlay"},
  {winter, orcish_fort}
)
human_castle = CT("human_castle", {"base", "overlay"}, {human, castle})
snow_human_castle = CT(
  "snow_human_castle",
  {"base", "overlay"},
  {frozen, human_castle}
)
elven_castle = CT("elven_castle", {"base", "overlay"}, {elven, castle})
dwarven_castle = CT("dwarven_castle", {"base", "overlay"}, {dwarven, castle})
desert_castle = CT("desert_castle", {"base", "overlay"}, {arid, castle})
ruined_human_castle = CT(
  "ruined_human_castle",
  {"base", "overlay"},
  {ruin, human_castle}
)
sunken_ruined_human_castle = CT(
  "sunken_ruined_human_castle",
  {"base", "overlay"},
  {ruined_human_castle, shallow_water}
)
swamp_ruined_human_castle = CT(
  "swamp_ruined_human_castle",
  {"base", "overlay"},
  {ruined_human_castle, swamp}
)

# Keeps:
human_bastion = CT("human_bastion", {"base", "overlay"}, {human, bastion})
snow_human_bastion = CT(
  "snow_human_bastion",
  {"base", "overlay"},
  {winter, human_bastion}
)
orcish_bastion = CT("orcish_bastion", {"base", "overlay"}, {orcish, bastion})
snow_orcish_bastion = CT(
  "snow_orcish_bastion",
  {"base", "overlay"},
  {winter, orcish_bastion}
)
human_keep = CT("human_keep", {"base", "overlay"}, {human, keep})
snow_human_keep = CT(
  "snow_human_keep",
  {"base", "overlay"},
  {winter, human_keep}
)
elven_keep = CT("elven_keep", {"base", "overlay"}, {elven, keep})
dwarven_keep = CT("dwarven_keep", {"base", "overlay"}, {dwarven, keep})
ruined_human_keep = CT(
  "ruined_human_keep",
  {"base", "overlay"},
  {ruin, human_keep}
)
sunken_ruined_human_keep = CT(
  "sunken_ruined_human_keep",
  {"base", "overlay"},
  {ruined_human_keep, shallow_water}
)
swamp_ruined_human_keep = CT(
  "swamp_ruined_human_keep",
  {"base", "overlay"},
  {ruined_human_keep, swamp}
)
desert_keep = CT("desert_keep", {"base", "overlay"}, {arid, keep})

# Improvements:
tracks_lr = CT("tracks_lr", {"overlay"}, {tracks, lr})
tracks_rl = CT("tracks_rl", {"overlay"}, {tracks, rl})
tracks_c = CT("tracks_c", {"overlay"}, {tracks, c})
wood_bridge_lr = CT("wood_bridge_lr", {"overlay"}, {wood_bridge, lr})
wood_bridge_rl = CT("wood_bridge_rl", {"overlay"}, {wood_bridge, rl})
wood_bridge_c = CT("wood_bridge_c", {"overlay"}, {wood_bridge, c})
stone_bridge_lr = CT("stone_bridge_lr", {"overlay"}, {stone_bridge, lr})
stone_bridge_rl = CT("stone_bridge_rl", {"overlay"}, {stone_bridge, rl})
stone_bridge_c = CT("stone_bridge_c", {"overlay"}, {stone_bridge, c})
chasm_bridge_lr = CT("chasm_bridge_lr", {"overlay"}, {chasm_bridge, lr})
chasm_bridge_rl = CT("chasm_bridge_rl", {"overlay"}, {chasm_bridge, rl})
chasm_bridge_c = CT("chasm_bridge_c", {"overlay"}, {chasm_bridge, c})



#############################################
# The mapping from map characters to types: #
#############################################

char_to_type = {
  'Wog': grey_deep_water,
  'Wo': deep_water,
  'Wot': tropical_deep_water,
  'Wwg': grey_shallow_water,
  'Ww': shallow_water,
  'Wwt': tropical_shallow_water,
  'Wwf': ford,
  'Wwr': reef,
  'Ss': swamp,
  'Sm': mud,
  'Gt': flat,
  'Gg': grass,
  'Gs': semi_dry_grass,
  'Gd': dry_grass,
  'Gll': leaf_litter,
  'Rb': dark_dirt,
  'Re': dirt,
  'Rd': dry_dirt,
  'Rr': cobbles,
  'Rrc': clean_cobbles,
  'Rp': overgrown_cobbles,
  'At': frozen,
  'Ai': ice,
  'Aa': snow,
  'Dd': desert,
  'Ds': sand,
  'Dd^Do': oasis,
  '^Dr': rubble,
  'Dd^Dc': crater,
  '^Efm': flowers,
  '^Gvs': crops,
  '^Es': stones,
  '^Em': mushrooms,
  '^Emf': mushroom_farm,
  '^Edp': desert_plants,
  '^Edpp': alt_desert_plants,
  '^Wm': windmill,
  '^Fet': great_tree,
  '^Ft': tropical_forest,
  '^Fp': evergreen_forest,
  '^Fpa': snow_evergreen_forest,
  '^Fds': deciduous_forest_summer,
  '^Fdf': deciduous_forest_fall,
  '^Fdw': deciduous_forest_winter,
  '^Fda': snow_deciduous_forest,
  '^Fms': mixed_forest_summer,
  '^Fmf': mixed_forest_fall,
  '^Fmw': mixed_forest_winter,
  '^Fma': snow_mixed_forest,
  'Hh': hills,
  'Hhd': dry_hills,
  'Hd': dunes,
  'Ha': snow_hills,
  'Mm': mountains,
  'Md': dry_mountains,
  'Ms': snow_mountains,
  'Iwr': wood_floor,
  '^Ii': light_beam,
  'Uu': cave_floor,
  'Uue': earthen_floor,
  'Urb': flagstones,
  'Ur': cave_path,
  '^Uf': mushroom_grove,
  '^Ufi': illuminated_mushroom_grove,
  'Uh': rockbound_cave,
  '^Br|': tracks_c,
  '^Br/': tracks_rl,
  '^Br\\': tracks_lr,
  'Qt': unwalkable,
  'Qxu': chasm,
  'Qxe': earthen_chasm,
  'Qxua': ethereal_abyss,
  'Ql': lava_chasm,
  'Qlf': lava,
  'Mv': volcano,
  'Mm^Xm': impassable_mountains,
  'Md^Xm': dry_impassable_mountains,
  'Ms^Xm': snow_impassable_mountains,
  'Xt': impassable,
  'Xu': cave_wall,
  'Xuc': hewn_cave_wall,
  'Xue': earthen_cave_wall,
  'Xuce': hewn_earthen_cave_wall,
  'Xos': stone_wall,
  'Xol': lit_stone_wall,
  '^Xo': impassibility,
  'Xv': void,
  'Vi': village,
  '^Vda': adobe_village,
  '^Vdt': desert_tents,
  '^Vct': tents,
  '^Vo': orcish_village,
  '^Voa': snow_orcish_village,
  '^Ve': elven_village,
  '^Vea': snow_elven_village,
  '^Vh': cottage,
  '^Vha': snow_cottage,
  '^Vhc': city_building,
  '^Vhca': snow_city_building,
  '^Vhh': hill_house,
  '^Vhha': snow_hill_house,
  '^Vht': tropical_village,
  '^Vd': drake_village,
  '^Vu': cave_village,
  '^Vud': dwarven_village,
  '^Vc': hut,
  '^Vca': snow_hut,
  '^Vl': log_cabin,
  '^Vla': snow_log_cabin,
  '^Vhs': swamp_village,
  'Ww^Vm': merfolk_village,
  'Ss^Vm': swamp_merfolk_village,
  'Ce': human_fort,
  'Cea': snow_human_fort,
  'Co': orcish_fort,
  'Coa': snow_orcish_fort,
  'Ch': human_castle,
  'Cha': snow_human_castle,
  'Cv': elven_castle,
  'Cud': dwarven_castle,
  'Chr': ruined_human_castle,
  'Chw': sunken_ruined_human_castle,
  'Chs': swamp_ruined_human_castle,
  'Cd': desert_castle,
  'Ke': human_bastion,
  'Kea': snow_human_bastion,
  'Ko': orcish_bastion,
  'Koa': snow_orcish_bastion,
  'Kh': human_keep,
  'Kha': snow_human_keep,
  'Kv': elven_keep,
  'Cud': dwarven_keep,
  'Khr': ruined_human_keep,
  'Khw': sunken_ruined_human_keep,
  'Khs': swamp_ruined_human_keep,
  'Kd': desert_keep,
  '^Cov': castle,
  '^Kov': keep,
  '^Bw|': wood_bridge_c,
  '^B/': wood_bridge_rl,
  '^B\\': wood_bridge_lr,
  '^Bsb|': stone_bridge_c,
  '^Bsb/': stone_bridge_rl,
  '^Bsb\\': stone_bridge_lr,
  '^Bs|': chasm_bridge_c,
  '^Bs/': chasm_bridge_rl,
  '^Bs\\': chasm_bridge_lr,
}

type_to_char = {type: char for (char, type) in char_to_type.items()}


######################
# Some useful stuff: #
######################

vars = dict(locals())
all_types = set()
for v in vars:
  if isinstance(vars[v], CT):
    all_types.add(vars[v])

movement_categories = { t for t in all_types if movement_category in t.parents }
conceptual_categories = {
  t for t in all_types if conceptual_category in t.parents
}
mechanics_categories = {
  t for t in all_types if mechanics_category in t.parents
}

movement_categories_of = {}
for t in all_types:
  movement_categories_of[t] = set()
  for a in t.ancestors():
    if a in movement_categories:
      movement_categories_of[t].add(a)

conceptual_categories_of = {}
for t in all_types:
  conceptual_categories_of[t] = set()
  for a in t.ancestors():
    if a in conceptual_categories:
      conceptual_categories_of[t].add(a)

mechanics_categories_of = {}
for t in all_types:
  mechanics_categories_of[t] = set()
  for a in t.ancestors():
    if a in mechanics_categories:
      mechanics_categories_of[t].add(a)

def optional(component):
  '''
  Returns a component that's a copy of the given one, except that it's required
  type is "any_type", meaning that as part of a chunk, it might be replaced by
  any other type when that chunk is added to another one.
  '''
  return C(component.real_type, component.position, any_type)

def with_required(component, typ):
  '''
  Returns a component that's a copy of the given one, except that it's required
  type is the given type, meaning that as part of a chunk, it might be replaced
  a component of the given type when that chunk is added to another one.
  '''
  return C(component.real_type, component.position, typ)
